# Allows to connect to GitLab with HTTPS
# https://github.com/JuliaDocs/Documenter.jl/issues/2061#issuecomment-1607077792
# https://github.com/JuliaDocs/Documenter.jl/blob/master/src/deployconfig.jl
using Documenter: DeployConfig, DeployDecision, marker, env_nonempty, HTTPS

@kwdef struct GitLabHTTPS <: DeployConfig
    commit_branch::String = get(ENV, "CI_COMMIT_BRANCH", "")
    pull_request_iid::String = get(ENV, "CI_EXTERNAL_PULL_REQUEST_IID", "")
    repo_path::String = get(ENV, "CI_PROJECT_PATH", "")
    repo_slug::String = get(ENV, "CI_PROJECT_PATH_SLUG", "")
    commit_tag::String = get(ENV, "CI_COMMIT_TAG", "")
    pipeline_source::String = get(ENV, "CI_PIPELINE_SOURCE", "")
end

Documenter.authentication_method(::GitLabHTTPS) = HTTPS

function Documenter.authenticated_repo_url(cfg::GitLabHTTPS) 
    token = get(ENV,"CI_BOT_TOKEN","")
    host = get(ENV,"CI_SERVER_HOST", "")
    return "https://documenter-ci:$token@$host/$(cfg.repo_path).git"
end

function Documenter.deploy_folder(
    cfg::GitLabHTTPS;
    repo,
    repo_previews = repo,
    devbranch,
    push_preview,
    devurl,
    branch = "docs",
    branch_previews = branch,
    tag_prefix = "",
    kwargs...,
)
    io = IOBuffer()
    all_ok = true

    println(io, "\nGitLab config:")
    println(io, "  Commit branch: \"", cfg.commit_branch, "\"")
    println(io, "  Pull request IID: \"", cfg.pull_request_iid, "\"")
    println(io, "  Repo slug: \"", cfg.repo_slug, "\"")
    println(io, "  Commit tag: \"", cfg.commit_tag, "\"")
    println(io, "  Pipeline source: \"", cfg.pipeline_source, "\"")

    build_type = if cfg.pull_request_iid != ""
        :preview
    elseif cfg.commit_tag != ""
        :release
    else
        :devbranch
    end

    println(io, "Detected build type: ", build_type)

    if build_type == :release
        tag_nobuild = version_tag_strip_build(cfg.commit_tag; tag_prefix)
        ## If a tag exist it should be a valid VersionNumber
        tag_ok = tag_nobuild !== nothing

        println(
            io,
            "- $(marker(tag_ok)) ENV[\"CI_COMMIT_TAG\"] contains a valid VersionNumber",
        )
        all_ok &= tag_ok

        is_preview = false
        subfolder = tag_nobuild
        deploy_branch = branch
        deploy_repo = repo
    elseif build_type == :preview
        pr_number = tryparse(Int, cfg.pull_request_iid)
        pr_ok = pr_number !== nothing
        all_ok &= pr_ok
        println(
            io,
            "- $(marker(pr_ok)) ENV[\"CI_EXTERNAL_PULL_REQUEST_IID\"]=\"$(cfg.pull_request_iid)\" is a number",
        )
        btype_ok = push_preview
        all_ok &= btype_ok
        is_preview = true
        println(
            io,
            "- $(marker(btype_ok)) `push_preview` keyword argument to deploydocs is `true`",
        )
        ## deploy to previews/PR
        subfolder = "previews/PR$(something(pr_number, 0))"
        deploy_branch = branch_previews
        deploy_repo = repo_previews
    else
        branch_ok = !isempty(cfg.commit_tag) || cfg.commit_branch == devbranch
        all_ok &= branch_ok
        println(
            io,
            "- $(marker(branch_ok)) ENV[\"CI_COMMIT_BRANCH\"] matches devbranch=\"$(devbranch)\"",
        )
        is_preview = false
        subfolder = devurl
        deploy_branch = branch
        deploy_repo = repo
    end

    key_ok = env_nonempty("CI_BOT_TOKEN")
    println(io, "- $(marker(key_ok)) ENV[\"CI_BOT_TOKEN\"] exists and is non-empty")
    all_ok &= key_ok

    print(io, "Deploying to folder $(repr(subfolder)): $(marker(all_ok))")
    @info String(take!(io))

    if all_ok
        return DeployDecision(;
            all_ok = true,
            branch = deploy_branch,
            repo = deploy_repo,
            subfolder = subfolder,
            is_preview = is_preview,
        )
    else
        return DeployDecision(; all_ok = false)
    end
end